require 'net/http'
require 'openssl'
require 'base64'


##
# Class HttpClient performs HTTP requests to the Secured Cloud API Server. 
# Authentication is handled via the supplied application key and shared secret.
#
# This class expects the pem file "/cert/sc.pem" to be found in the working 
# directory of the application using this library. This pem file is used for
# SSL validation.
#
# @author::  Alan Vella
##

class HttpClient


  APP_CONTENT_TYPE = "application/vnd.securedcloud.v7.0+json"
  AUTH_SCHEME = "SC "
  PEM_FILE_PATH = Dir.pwd + "/cert/sc.pem"
  
  
  #Run checks before issuing request.
  def self.runChecks
    if !(File.exists?(PEM_FILE_PATH)) then
      raise "A valid pem file named sc.pem is required for SSL validation. This file must be placed in the cert directory of your vagrant workspace."
    end
  end
  
  
  
  #Raise error if response is not successful.
  def self.handleResponseError(response)
    
    if ((response.code != "202") && (response.code != "200")) then
      
      errorMsg = ""
      
      httpStatusDesc = ""
      
      #Add description of HTTP error code
      if (response.code == "400") then
        httpStatusDesc = "Bad Request"
        
      elsif (response.code == "403") then
        httpStatusDesc = "Forbidden"
        
      elsif (response.code == "404") then
        httpStatusDesc = "Resource Not Found"
        
      elsif (response.code == "405") then
        httpStatusDesc = "Method Not Allowed"
        
      elsif (response.code == "405") then
        httpStatusDesc = "Too Many Requests"
        
      elsif (response.code == "500") then
        httpStatusDesc = "Internal Error"
        
      end
      
      #Add error message from SecuredCloud API if there is any
      if ((response['X-Application-Error-Reference'] != nil) && (response['X-Application-Error-Reference'] != "")) then
        errorMsg = response['X-Application-Error-Reference']
      end
      if ((response['X-Application-Error-Description'] != nil) && (response['X-Application-Error-Description'] != "")) then
        errorMsg = response['X-Application-Error-Description']
      end
      
      #Build final error message to be sent to user.
      if (errorMsg == "") then
        errorMsg = "Error " + response.code.to_str() + " (" + httpStatusDesc + ") " + response.body()
      else
        errorMsg = "Error " + response.code.to_str() + " (" + httpStatusDesc + ") " + errorMsg  
      end
      
      raise errorMsg
    end
  end


  #Sends an HTTP GET request and returns response.
  def self.sendGETRequest(authInfo, url)
  
    self.runChecks
  
    begin
      
      @applicationKey = authInfo.getApplicationKey()
      @sharedSecret = authInfo.getSharedSecret()
    
      @url = URI.parse(url)
        
      #Setup SSL validation.
      http = Net::HTTP.new(@url.host, @url.port)
      http.use_ssl = true
      http.verify_mode = OpenSSL::SSL::VERIFY_PEER
      http.cert_store = OpenSSL::X509::Store.new
      http.cert_store.set_default_paths
      http.cert_store.add_file(PEM_FILE_PATH)
       
      #Create and execute request
      request = Net::HTTP::Get.new(@url.request_uri)
      request['Accept'] = APP_CONTENT_TYPE  
      request['Authorization'] = AUTH_SCHEME + createAuthorization("GET", url)
      response = http.request(request)
      
      rescue Exception => e  
        raise "Error creating API request: " + e.message
      
    end    
    
    #Raise error if response is not successful.
    self.handleResponseError(response)
    
    #Return response.
    return response
  end

  

  
  #Sends an HTTP POST request and returns response.
  def self.sendPOSTRequest(authInfo, url, body)
    
    self.runChecks
    
    begin
    
      @applicationKey = authInfo.getApplicationKey()
      @sharedSecret = authInfo.getSharedSecret()
    
      @url = URI.parse(url)
        
      #Setup SSL validation.
      http = Net::HTTP.new(@url.host, @url.port)
      http.use_ssl = true
      http.verify_mode = OpenSSL::SSL::VERIFY_PEER
      http.cert_store = OpenSSL::X509::Store.new
      http.cert_store.set_default_paths
      http.cert_store.add_file(PEM_FILE_PATH)
       
      #Create and execute request
      request = Net::HTTP::Post.new(@url.request_uri)
      request['Accept'] = APP_CONTENT_TYPE  
      request['Content-Type'] = APP_CONTENT_TYPE
      request['Authorization'] = AUTH_SCHEME + createAuthorization("POST", url)
      request.body = body
      response = http.request(request)
      
      rescue Exception => e  
        raise "Error creating API request: " + e.message    
    
    end
    
    #Raise error if response is not successful.
    self.handleResponseError(response)
    
    #Return response.
    return response
  end
  

  #Sends an HTTP DELETE request and returns response.
  def self.sendDELETERequest(authInfo, url, params = nil)
    
    self.runChecks
    
    begin
    
      @applicationKey = authInfo.getApplicationKey()
      @sharedSecret = authInfo.getSharedSecret()
    
      #Add parameters if present
      if (params != nil) then
        url += "?"
        params.each do |param|
          url += param[0] + "=" + param[1] + "&"
        end
      
        #Remove last ampersand
        url = url[0...-1]
      end
    
      @url = URI.parse(url)
        
      #Setup SSL validation.
      http = Net::HTTP.new(@url.host, @url.port)
      http.use_ssl = true
      http.verify_mode = OpenSSL::SSL::VERIFY_PEER
      http.cert_store = OpenSSL::X509::Store.new
      http.cert_store.set_default_paths
      http.cert_store.add_file(PEM_FILE_PATH)
       
      #Create and execute request
      request = Net::HTTP::Delete.new(@url.request_uri)
      request['Accept'] = APP_CONTENT_TYPE  
      request['Content-Type'] = APP_CONTENT_TYPE
      request['Authorization'] = AUTH_SCHEME + createAuthorization("DELETE", url)
      response = http.request(request)  
      
      rescue Exception => e  
        raise "Error creating API request: " + e.message
    
    end  
    
    #Raise error if response is not successful.
    self.handleResponseError(response)
    
    #Return response.
    return response
    
  end
       
    
  #Sends an HTTP PUT request and returns response.
  def self.sendPUTRequest(authInfo, url, params = nil)
    
    self.runChecks
    
    begin
    
      @applicationKey = authInfo.getApplicationKey()
      @sharedSecret = authInfo.getSharedSecret()
    
      #Add parameters if present
      if (params != nil) then
        url += "?"
        params.each do |param|
          url += param[0] + "=" + param[1] + "&"
        end
      
        #Remove last ampersand
        url = url[0...-1]
      end
    
      @url = URI.parse(url)
        
      #Setup SSL validation.
      http = Net::HTTP.new(@url.host, @url.port)
      http.use_ssl = true
      http.verify_mode = OpenSSL::SSL::VERIFY_PEER
      http.cert_store = OpenSSL::X509::Store.new
      http.cert_store.set_default_paths
      http.cert_store.add_file(PEM_FILE_PATH)
       
      #Create and execute request
      request = Net::HTTP::Put.new(@url.request_uri)
      request['Accept'] = APP_CONTENT_TYPE  
      request['Content-Type'] = APP_CONTENT_TYPE
      request['Authorization'] = AUTH_SCHEME + createAuthorization("PUT", url)
      response = http.request(request)
      
      rescue Exception => e  
        raise "Error creating API request: " + e.message
        
    end    
    
    #Raise error if response is not successful.
    self.handleResponseError(response)
    
    #Return response.
    return response
    
  end


  #Removes the servlet name from a given external API URL.
  def self.removeServletName(url)
    urlAsArray = url.split("/");
    newURL = ""
    for i in 4..(urlAsArray.size() - 1)
      newURL += "/" + urlAsArray[i]
    end
    return newURL
  end


  #Creates authorization header for a given method and URL.
  def self.createAuthorization(method, url)
    stringToSign = method + " " + removeServletName(url) + " " + @applicationKey
    hash = OpenSSL::HMAC.digest('sha256', @sharedSecret, stringToSign)
    requestSignature = Base64.encode64(hash).gsub(/\s+/, ' ').strip    
    credentials = @applicationKey + ":" + requestSignature;    
    encodedCredentials = Base64.encode64(credentials).gsub(/\s+/, ' ').strip
    return encodedCredentials
  end
end
